/* BLURB gpl

                           Coda File System
                              Release 5

          Copyright (c) 1987-1999 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

                        Additional copyrights
                           none currently

#*/

#ifdef __cplusplus
extern "C" {
#endif __cplusplus

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* System-wide include files */
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <time.h>
#include <stdio.h>
#include "coda_string.h"
#include <errno.h>
#include <stdlib.h>
#include <netdb.h>

#ifdef __cplusplus
}
#endif __cplusplus


/* Coda include files */
#include <util.h>

/* Local include files */
#include "advice_srv.h"
#include "data.h"
#include "globals.h"



/* Venus Version Information */
int VenusMajorVersionNumber;
int VenusMinorVersionNumber;

/* Command Line Arguments */
int debug = 0;
int verbose = 0;
int AutoReply = 0;               /* -a will set (or unset) this value */ 
int LogLevel = 0;                /* -d <int> */
int StackChecking = 0;		 /* -s */


/* Logging Information */
FILE *LogFile;
char LogFileName[MAXPATHLEN];       /* DEFAULT: /coda/usr/<username>/.questionnaires/advice.log */
FILE *EventFile;
char EventFileName[MAXPATHLEN];     /* DEFAULT: /coda/usr/<username>/.questionnaires/advice.event */

/* Log Levels */
int EnterLeaveMsgs = 100;

/* Storage Information */
char BaseDir[MAXPATHLEN];
char WorkingDir[MAXPATHLEN];
char tmpDir[MAXPATHLEN];

/* Error Information */
char error_msg[BUFSIZ];

/* Process Information */
int thisPGID;
int thisPID;
int childPID;
int interfacePID = -1;

/* User Information */
char UserName[16];
int uid = -1;

/* Host Information */
char ShortHostName[MAXHOSTNAMELEN];
char HostName[MAXHOSTNAMELEN];

/* Connection Information */
int WeLostTheConnection = FALSE;

/* Synchronization Variables */
char shutdownSync;
char dataSync;
char programlogSync;
char replacementlogSync;
char initialUserSync;
char userSync;
char workerSync;

char discomissSync;
char weakmissSync;
char readmissSync;
char miscSync;
char hoardwalkSync;

/* Return Answer Variables */
int weakmissAnswer;
int readmissAnswer;

/* RPC Parameter Values */
char ProgramAccessLog[MAXPATHLEN];
char ReplacementLog[MAXPATHLEN];

#ifndef HAVE_SETENV
/* setenv replacement ... uses putenv */
static  int
setenv(const char *name, const char *value, int overwrite)
{
  char string[MAXPATHLEN];
  snprintf (string, MAXPATHLEN, "%s=%s", name, value);
  return putenv(string);
}
#endif


/* Command-Line Argument Routines */
void PrintUsage() {
  fprintf(stderr, "advice_srv [-d <LogLevel>] [i] [-v]\n");
  fflush(stderr);
  exit(-1);
}

/*PRIVATE*/ 
void CommandLineArgs(int argc, char **argv) {
  extern int optind;
  int arg_counter = 1;

  while ((arg_counter = getopt(argc, argv, "ad:im:svw:")) != EOF)
    switch(arg_counter) {
      case 'a':
	      // Undocumented feature 
	      // DEFAULT is ON (until PCM is functional)
              fprintf(stderr, "AutoReply turned on.\n");
              AutoReply = 1;
              break;
      case 'd':
              debug = 1;
	      LogLevel = atoi(argv[optind-1]);
              break;
      case 'i':
	      InitiateNewUser();
	      break;
      case 'm':
	      extern struct timeval cont_sw_threshold;
	      cont_sw_threshold.tv_sec = atoi(argv[optind-1]);
	      cont_sw_threshold.tv_usec = 0;
	      break;
      case 's':
	      StackChecking = 1;
	      break;
      case 'v':
              verbose = 1;
              break;
      case 'w':
	      extern struct timeval run_wait_threshold;
	      run_wait_threshold.tv_sec = (atoi(argv[optind-1]));
	      run_wait_threshold.tv_usec = 0;
	      break;
      default:
              PrintUsage();
    }
}


void InitiateNewUser()
{
  /* Give them some sort of tutorial?? */
  fprintf(stderr, "Tutorial for new users doesn't exist!  Bug Maria...\n");

  /* Create a unique Coda directory in which to dump our data */
  /* dflt: /coda/usr/<username>/.questionnaires */
  InitializeDirectory(BaseDir, UserName);

  if (strncmp(BaseDir, "/coda", 5) == 0)
      SetAccessLists(BaseDir, UserName);

  fprintf(stderr, "New user (%s) initiated.\n", UserName);
}


/* Initialization Routines */

void InitLogFile() {
  // LogFileName: 
  //    /coda/usr/<username>/.questionnaires/<HOSTNAME>_<PID>_<DATE:TIME>/advice.log  
    snprintf(LogFileName, MAXPATHLEN, "%s/advice.log", WorkingDir);

    LogFile = fopen(LogFileName, "a");
    if (LogFile == NULL) {
        fprintf(stderr, "LOGFILE (%s) initialization failed (errno=%d)\n\n", LogFileName, errno);
	fflush(stderr);
    }
    //  Use the following line to send the LogFile output to stdout...
    //    LogFile = stdout;

    struct timeval now;
    gettimeofday(&now, 0);
    char *s = ctime((const long int *)&now.tv_sec);
    LogMsg(0,LogLevel,LogFile,"LOGFILE initialized with LogLevel = %d at %s",
           LogLevel,ctime((const long int *)&now.tv_sec));
    LogMsg(0,LogLevel,LogFile,"My pid is %d",getpid());
}

void InitEventFile() {
  // EventFileName:
  //   /coda/usr/<username>/.questionnaires/<HOSTNAME>_<PID>_<DATE:TIME>/advice.event  
    snprintf(EventFileName, MAXPATHLEN, "%s/advice.event", WorkingDir);

    EventFile = fopen(EventFileName, "a");
    if (LogFile == NULL) 
        LogMsg(0,LogLevel,LogFile, "EVENTFILE (%s) initialization failed (errno=%d)", EventFileName, errno);
}

void InitPGID() {
#ifdef SETPGRP_VOID
        (void) setpgrp();
#else
        (void) setpgrp(0, thisPID);
#endif
  thisPGID = thisPID;
}

void InitEnvironment() {
    int error = 0;

    error += setenv("TCL_LIBRARY", TCL, 1);
    error += setenv("TK_LIBRARY", TK, 1);
    error += setenv("TIX_LIBRARY", TIX, 1);
    CODA_ASSERT(error == 0);
}

void InitHostName() {
  CODA_ASSERT(gethostname(HostName, MAXHOSTNAMELEN) == 0);
  strcpy(ShortHostName, HostName);
  for (int i = 0; i < strlen(HostName); i++)
	  if (ShortHostName[i] == '.')
		  ShortHostName[i] = 0;
  snprintf(HostName, MAXHOSTNAMELEN, "localhost");
}


